from ctypes import *

isl = cdll.LoadLibrary("libisl.so")

class Context:
  defaultInstance = None
  instances = {}

  def __init__(self):
    ptr = isl.isl_ctx_alloc()
    self.ptr = ptr
    Context.instances[ptr] = self

  def __del__(self):
    isl.isl_ctx_free(self)

  def from_param(self):
    return self.ptr

  @staticmethod
  def from_ptr(ptr):
    return Context.instances[ptr]

  @staticmethod
  def getDefaultInstance():
    if Context.defaultInstance == None:
      Context.defaultInstance = Context()

    return Context.defaultInstance

class IslObject:
  def __init__(self, string = "", ctx = None, ptr = None):
    self.initialize_isl_methods()
    if ptr != None:
      self.ptr = ptr
      self.ctx = self.get_isl_method("get_ctx")(self)
      return

    if ctx == None:
      ctx = Context.getDefaultInstance()

    self.ctx = ctx
    self.ptr = self.get_isl_method("read_from_str")(ctx, string, -1)

  def __del__(self):
    self.get_isl_method("free")(self)

  def from_param(self):
    return self.ptr

  @property
  def context(self):
    return self.ctx

  def __repr__(self):
    p = Printer(self.ctx)
    self.to_printer(p)
    return p.getString();

  def __str__(self):
    p = Printer(self.ctx)
    self.to_printer(p)
    return p.getString();

  @staticmethod
  def isl_name():
    return "No isl name available"

  def initialize_isl_methods(self):
    if hasattr(self.__class__, "initialized"):
      return

    self.__class__.initalized = True
    self.get_isl_method("read_from_str").argtypes = [Context, c_char_p, c_int]
    self.get_isl_method("copy").argtypes = [self.__class__]
    self.get_isl_method("copy").restype = c_int
    self.get_isl_method("free").argtypes = [self.__class__]
    self.get_isl_method("get_ctx").argtypes = [self.__class__]
    self.get_isl_method("get_ctx").restype = Context.from_ptr 
    getattr(isl, "isl_printer_print_" + self.isl_name()).argtypes = [Printer, self.__class__]

  def get_isl_method(self, name):
    return getattr(isl, "isl_" + self.isl_name() + "_" + name)

  def to_printer(self, printer):
    getattr(isl, "isl_printer_print_" + self.isl_name())(printer, self)

class BSet(IslObject):
  @staticmethod
  def from_ptr(ptr):
    if not ptr:
      return
    return BSet(ptr = ptr)

  @staticmethod
  def isl_name():
    return "basic_set"

class Set(IslObject):
  @staticmethod
  def from_ptr(ptr):
    if not ptr:
      return
    return Set(ptr = ptr)

  @staticmethod
  def isl_name():
    return "set"

class USet(IslObject):
  @staticmethod
  def from_ptr(ptr):
    if not ptr:
      return
    return USet(ptr = ptr)

  @staticmethod
  def isl_name():
    return "union_set"


class BMap(IslObject):
  @staticmethod
  def from_ptr(ptr):
    if not ptr:
      return
    return BMap(ptr = ptr)

  def __mul__(self, set):
    return self.intersect_domain(set)

  @staticmethod
  def isl_name():
    return "basic_map"

class Map(IslObject):
  @staticmethod
  def from_ptr(ptr):
    if not ptr:
      return
    return Map(ptr = ptr)

  def __mul__(self, set):
    return self.intersect_domain(set)

  @staticmethod
  def isl_name():
    return "map"

  @staticmethod
  def lex_lt(dim):
    dim = isl.isl_dim_copy(dim)
    return isl.isl_map_lex_lt(dim)

  @staticmethod
  def lex_le(dim):
    dim = isl.isl_dim_copy(dim)
    return isl.isl_map_lex_le(dim)

  @staticmethod
  def lex_gt(dim):
    dim = isl.isl_dim_copy(dim)
    return isl.isl_map_lex_gt(dim)

  @staticmethod
  def lex_ge(dim):
    dim = isl.isl_dim_copy(dim)
    return isl.isl_map_lex_ge(dim)

class UMap(IslObject):
  @staticmethod
  def from_ptr(ptr):
    if not ptr:
      return
    return UMap(ptr = ptr)

  @staticmethod
  def isl_name():
    return "union_map"

class Dim(IslObject):
  @staticmethod
  def from_ptr(ptr):
    if not ptr:
      return
    return Dim(ptr = ptr)

  @staticmethod
  def isl_name():
    return "dim"

  def initialize_isl_methods(self):
    if hasattr(self.__class__, "initialized"):
      return

    self.__class__.initalized = True
    self.get_isl_method("copy").argtypes = [self.__class__]
    self.get_isl_method("copy").restype = c_int
    self.get_isl_method("free").argtypes = [self.__class__]
    self.get_isl_method("get_ctx").argtypes = [self.__class__]
    self.get_isl_method("get_ctx").restype = Context.from_ptr 

  def __repr__(self):
    return str(self)

  def __str__(self):

    dimParam = isl.isl_dim_size(self, 1)
    dimIn = isl.isl_dim_size(self, 2)
    dimOut = isl.isl_dim_size(self, 3)

    if dimIn:
      return "<dim In:%s, Out:%s, Param:%s>" % (dimIn, dimOut, dimParam)

    return "<dim Set:%s, Param:%s>" % (dimOut, dimParam)

class Printer:
  FORMAT_ISL = 0
  FORMAT_POLYLIB = 1
  FORMAT_POLYLIB_CONSTRAINTS = 2
  FORMAT_OMEGA = 3
  FORMAT_C = 4
  FORMAT_LATEX = 5
  FORMAT_EXT_POLYLIB = 6

  def __init__(self, ctx = None):
    if ctx == None:
      ctx = Context.getDefaultInstance()

    self.ctx = ctx
    self.ptr = isl.isl_printer_to_str(ctx)

  def setFormat(self, format):
    self.ptr = isl.isl_printer_set_output_format(self, format);

  def from_param(self):
    return self.ptr

  def __del__(self):
    isl.isl_printer_free(self)

  def getString(self):
    return isl.isl_printer_get_str(self)

functions = [
             # Unary properties
             ("is_empty", BSet, [BSet], c_int),
             ("is_empty", Set, [Set], c_int),
             ("is_empty", USet, [USet], c_int),
             ("is_empty", BMap, [BMap], c_int),
             ("is_empty", Map, [Map], c_int),
             ("is_empty", UMap, [UMap], c_int),

    #         ("is_universe", Set, [Set], c_int),
    #         ("is_universe", Map, [Map], c_int),

             ("is_single_valued", Map, [Map], c_int),

             ("is_bijective", Map, [Map], c_int),

             ("is_wrapping", BSet, [BSet], c_int),
             ("is_wrapping", Set, [Set], c_int),

             # Binary properties
             ("is_equal", BSet, [BSet, BSet], c_int),
             ("is_equal", Set, [Set, Set], c_int),
             ("is_equal", USet, [USet, USet], c_int),
             ("is_equal", BMap, [BMap, BMap], c_int),
             ("is_equal", Map, [Map, Map], c_int),
             ("is_equal", UMap, [UMap, UMap], c_int),

             # is_disjoint missing

             # ("is_subset", BSet, [BSet, BSet], c_int),
             ("is_subset", Set, [Set, Set], c_int),
             ("is_subset", USet, [USet, USet], c_int),
             ("is_subset", BMap, [BMap, BMap], c_int),
             ("is_subset", Map, [Map, Map], c_int),
             ("is_subset", UMap, [UMap, UMap], c_int),
             #("is_strict_subset", BSet, [BSet, BSet], c_int),
             ("is_strict_subset", Set, [Set, Set], c_int),
             ("is_strict_subset", USet, [USet, USet], c_int),
             ("is_strict_subset", BMap, [BMap, BMap], c_int),
             ("is_strict_subset", Map, [Map, Map], c_int),
             ("is_strict_subset", UMap, [UMap, UMap], c_int),

             # Unary Operations
             ("complement", Set, [Set], Set),
             ("reverse", BMap, [BMap], BMap),
             ("reverse", Map, [Map], Map),
             ("reverse", UMap, [UMap], UMap),

             # Projection missing
             ("range", BMap, [BMap], BSet),
             ("range", Map, [Map], Set),
             ("range", UMap, [UMap], USet),
             ("domain", BMap, [BMap], BSet),
             ("domain", Map, [Map], Set),
             ("domain", UMap, [UMap], USet),

             ("identity", Set, [Set], Map),
             ("identity", USet, [USet], UMap),

             ("deltas", BMap, [BMap], BSet),
             ("deltas", Map, [Map], Set),
             ("deltas", UMap, [UMap], USet),

             ("coalesce", Set, [Set], Set),
             ("coalesce", USet, [USet], USet),
             ("coalesce", Map, [Map], Map),
             ("coalesce", UMap, [UMap], UMap),

             ("detect_equalities", BSet, [BSet], BSet),
             ("detect_equalities", Set, [Set], Set),
             ("detect_equalities", USet, [USet], USet),
             ("detect_equalities", BMap, [BMap], BMap),
             ("detect_equalities", Map, [Map], Map),
             ("detect_equalities", UMap, [UMap], UMap),

             ("convex_hull", Set, [Set], Set),
             ("convex_hull", Map, [Map], Map),

             ("simple_hull", Set, [Set], Set),
             ("simple_hull", Map, [Map], Map),

             ("affine_hull", BSet, [BSet], BSet),
             ("affine_hull", Set, [Set], BSet),
             ("affine_hull", USet, [USet], USet),
             ("affine_hull", BMap, [BMap], BMap),
             ("affine_hull", Map, [Map], BMap),
             ("affine_hull", UMap, [UMap], UMap),

             ("polyhedral_hull", Set, [Set], Set),
             ("polyhedral_hull", USet, [USet], USet),
             ("polyhedral_hull", Map, [Map], Map),
             ("polyhedral_hull", UMap, [UMap], UMap),

             # Power missing
             # Transitive closure missing
             # Reaching path lengths missing

             ("wrap", BMap, [BMap], BSet),
             ("wrap", Map, [Map], Set),
             ("wrap", UMap, [UMap], USet),
             ("unwrap", BSet, [BMap], BMap),
             ("unwrap", Set, [Map], Map),
             ("unwrap", USet, [UMap], UMap),

             ("flatten", Set, [Set], Set),
             ("flatten", Map, [Map], Map),
             ("flatten_map", Set, [Set], Map),

             # Dimension manipulation missing

             # Binary Operations
             ("intersect", BSet, [BSet, BSet], BSet),
             ("intersect", Set, [Set, Set], Set),
             ("intersect", USet, [USet, USet], USet),
             ("intersect", BMap, [BMap, BMap], BMap),
             ("intersect", Map, [Map, Map], Map),
             ("intersect", UMap, [UMap, UMap], UMap),
             ("intersect_domain", BMap, [BMap, BSet], BMap),
             ("intersect_domain", Map, [Map, Set], Map),
             ("intersect_domain", UMap, [UMap, USet], UMap),
             ("intersect_range", BMap, [BMap, BSet], BMap),
             ("intersect_range", Map, [Map, Set], Map),
             ("intersect_range", UMap, [UMap, USet], UMap),

             ("union", BSet, [BSet, BSet], Set),
             ("union", Set, [Set, Set], Set),
             ("union", USet, [USet, USet], USet),
             ("union", BMap, [BMap, BMap], Map),
             ("union", Map, [Map, Map], Map),
             ("union", UMap, [UMap, UMap], UMap),

             ("subtract", Set, [Set, Set], Set),
             ("subtract", Map, [Map, Map], Map),
             ("subtract", USet, [USet, USet], USet),
             ("subtract", UMap, [UMap, UMap], UMap),

             ("apply", BSet, [BSet, BMap], BSet),
             ("apply", Set, [Set, Map], Set),
             ("apply", USet, [USet, UMap], USet),
             ("apply_domain", BMap, [BMap, BMap], BMap),
             ("apply_domain", Map, [Map, Map], Map),
             ("apply_domain", UMap, [UMap, UMap], UMap),
             ("apply_range", BMap, [BMap, BMap], BMap),
             ("apply_range", Map, [Map, Map], Map),
             ("apply_range", UMap, [UMap, UMap], UMap),

             ("gist", BSet, [BSet, BSet], BSet),
             ("gist", Set, [Set, Set], Set),
             ("gist", USet, [USet, USet], USet),
             ("gist", BMap, [BMap, BMap], BMap),
             ("gist", Map, [Map, Map], Map),
             ("gist", UMap, [UMap, UMap], UMap),

             # Lexicographic Optimizations
             # partial_lexmin missing
             ("lexmin", BSet, [BSet], BSet),
             ("lexmin", Set, [Set], Set),
             ("lexmin", USet, [USet], USet),
             ("lexmin", BMap, [BMap], BMap),
             ("lexmin", Map, [Map], Map),
             ("lexmin", UMap, [UMap], UMap),

             ("lexmax", BSet, [BSet], BSet),
             ("lexmax", Set, [Set], Set),
             ("lexmax", USet, [USet], USet),
             ("lexmax", BMap, [BMap], BMap),
             ("lexmax", Map, [Map], Map),
             ("lexmax", UMap, [UMap], UMap),

              # Undocumented
             ("lex_lt_union_set", USet, [USet, USet], UMap),
             ("lex_le_union_set", USet, [USet, USet], UMap),
             ("lex_gt_union_set", USet, [USet, USet], UMap),
             ("lex_ge_union_set", USet, [USet, USet], UMap),

             ]
keep_functions = [
             # Unary properties
             ("get_dim", BSet, [BSet], Dim),
             ("get_dim", Set, [Set], Dim),
             ("get_dim", USet, [USet], Dim),
             ("get_dim", BMap, [BMap], Dim),
             ("get_dim", Map, [Map], Dim),
             ("get_dim", UMap, [UMap], Dim)
             ]

def addIslFunction(object, name):
    functionName = "isl_" + object.isl_name() + "_" + name
    islFunction = getattr(isl, functionName)
    if len(islFunction.argtypes) == 1:
      f = lambda a: islFunctionOneOp(islFunction, a)
    elif len(islFunction.argtypes) == 2:
      f = lambda a, b: islFunctionTwoOp(islFunction, a, b)
    object.__dict__[name] = f


def islFunctionOneOp(islFunction, ops):
  ops = getattr(isl, "isl_" + ops.isl_name() + "_copy")(ops)
  return islFunction(ops)

def islFunctionTwoOp(islFunction, opOne, opTwo):
  opOne = getattr(isl, "isl_" + opOne.isl_name() + "_copy")(opOne)
  opTwo = getattr(isl, "isl_" + opTwo.isl_name() + "_copy")(opTwo)
  return islFunction(opOne, opTwo)

for (operation, base, operands, ret) in functions:
  functionName = "isl_" + base.isl_name() + "_" + operation
  islFunction = getattr(isl, functionName)
  if len(operands) == 1:
    islFunction.argtypes = [c_int]
  elif len(operands) == 2:
    islFunction.argtypes = [c_int, c_int]

  if ret == c_int:
    islFunction.restype = ret
  else:
    islFunction.restype = ret.from_ptr

  addIslFunction(base, operation)

def addIslFunctionKeep(object, name):
    functionName = "isl_" + object.isl_name() + "_" + name
    islFunction = getattr(isl, functionName)
    if len(islFunction.argtypes) == 1:
      f = lambda a: islFunctionOneOpKeep(islFunction, a)
    elif len(islFunction.argtypes) == 2:
      f = lambda a, b: islFunctionTwoOpKeep(islFunction, a, b)
    object.__dict__[name] = f

def islFunctionOneOpKeep(islFunction, ops):
  return islFunction(ops)

def islFunctionTwoOpKeep(islFunction, opOne, opTwo):
  return islFunction(opOne, opTwo)

for (operation, base, operands, ret) in keep_functions:
  functionName = "isl_" + base.isl_name() + "_" + operation
  islFunction = getattr(isl, functionName)
  if len(operands) == 1:
    islFunction.argtypes = [c_int]
  elif len(operands) == 2:
    islFunction.argtypes = [c_int, c_int]

  if ret == c_int:
    islFunction.restype = ret
  else:
    islFunction.restype = ret.from_ptr

  addIslFunctionKeep(base, operation)

isl.isl_ctx_free.argtypes = [Context]
isl.isl_basic_set_read_from_str.argtypes = [Context, c_char_p, c_int]
isl.isl_set_read_from_str.argtypes = [Context, c_char_p, c_int]
isl.isl_basic_set_copy.argtypes = [BSet]
isl.isl_basic_set_copy.restype = c_int
isl.isl_set_copy.argtypes = [Set]
isl.isl_set_copy.restype = c_int
isl.isl_set_copy.argtypes = [Set]
isl.isl_set_copy.restype = c_int
isl.isl_set_free.argtypes = [Set]
isl.isl_basic_set_get_ctx.argtypes = [BSet]
isl.isl_basic_set_get_ctx.restype = Context.from_ptr
isl.isl_set_get_ctx.argtypes = [Set]
isl.isl_set_get_ctx.restype = Context.from_ptr
isl.isl_basic_set_get_dim.argtypes = [BSet]
isl.isl_basic_set_get_dim.restype = Dim.from_ptr
isl.isl_set_get_dim.argtypes = [Set]
isl.isl_set_get_dim.restype = Dim.from_ptr
isl.isl_union_set_get_dim.argtypes = [USet]
isl.isl_union_set_get_dim.restype = Dim.from_ptr

isl.isl_basic_map_read_from_str.argtypes = [Context, c_char_p, c_int]
isl.isl_map_read_from_str.argtypes = [Context, c_char_p, c_int]
isl.isl_basic_map_free.argtypes = [BMap]
isl.isl_map_free.argtypes = [Map]
isl.isl_basic_map_copy.argtypes = [BMap]
isl.isl_basic_map_copy.restype = c_int
isl.isl_map_copy.argtypes = [Map]
isl.isl_map_copy.restype = c_int
isl.isl_map_get_ctx.argtypes = [Map]
isl.isl_basic_map_get_ctx.argtypes = [BMap]
isl.isl_basic_map_get_ctx.restype = Context.from_ptr
isl.isl_map_get_ctx.argtypes = [Map]
isl.isl_map_get_ctx.restype = Context.from_ptr
isl.isl_basic_map_get_dim.argtypes = [BMap]
isl.isl_basic_map_get_dim.restype = Dim.from_ptr
isl.isl_map_get_dim.argtypes = [Map]
isl.isl_map_get_dim.restype = Dim.from_ptr
isl.isl_union_map_get_dim.argtypes = [UMap]
isl.isl_union_map_get_dim.restype = Dim.from_ptr
isl.isl_printer_free.argtypes = [Printer]
isl.isl_printer_to_str.argtypes = [Context]
isl.isl_printer_print_basic_set.argtypes = [Printer, BSet]
isl.isl_printer_print_set.argtypes = [Printer, Set]
isl.isl_printer_print_basic_map.argtypes = [Printer, BMap]
isl.isl_printer_print_map.argtypes = [Printer, Map]
isl.isl_printer_get_str.argtypes = [Printer]
isl.isl_printer_get_str.restype = c_char_p
isl.isl_printer_set_output_format.argtypes = [Printer, c_int]
isl.isl_printer_set_output_format.restype = c_int
isl.isl_dim_size.argtypes = [Dim, c_int]
isl.isl_dim_size.restype = c_int

isl.isl_map_lex_lt.argtypes = [c_int]
isl.isl_map_lex_lt.restype = Map.from_ptr
isl.isl_map_lex_le.argtypes = [c_int]
isl.isl_map_lex_le.restype = Map.from_ptr
isl.isl_map_lex_gt.argtypes = [c_int]
isl.isl_map_lex_gt.restype = Map.from_ptr
isl.isl_map_lex_ge.argtypes = [c_int]
isl.isl_map_lex_ge.restype = Map.from_ptr

isl.isl_union_map_compute_flow.argtypes = [c_int, c_int, c_int, c_int, c_void_p,
                                           c_void_p, c_void_p, c_void_p]

def dependences(sink, must_source, may_source, schedule):
  sink = getattr(isl, "isl_" + sink.isl_name() + "_copy")(sink)
  must_source = getattr(isl, "isl_" + must_source.isl_name() + "_copy")(must_source)
  may_source = getattr(isl, "isl_" + may_source.isl_name() + "_copy")(may_source)
  schedule = getattr(isl, "isl_" + schedule.isl_name() + "_copy")(schedule)
  must_dep = c_int()
  may_dep = c_int()
  must_no_source = c_int()
  may_no_source = c_int()
  isl.isl_union_map_compute_flow(sink, must_source, may_source, schedule, \
                                 byref(must_dep), byref(may_dep),
                                 byref(must_no_source),
                                 byref(may_no_source))

  return (UMap.from_ptr(must_dep), UMap.from_ptr(may_dep), \
          USet.from_ptr(must_no_source), USet.from_ptr(may_no_source))


__all__ = ['Set', 'Map', 'Printer', 'Context']



import gdb
import re

# GDB Pretty Printers for most isl objects
class IslObjectPrinter:
	"""Print an isl object"""
	def __init__ (self, val, type):
                self.val = val
                self.type = type

	def to_string (self):
                # Cast val to a void pointer to stop gdb using this pretty
                # printer for the pointer which would lead to an infinite loop.
                void_ptr = gdb.lookup_type('void').pointer()
                value = str(self.val.cast(void_ptr))
                printer = gdb.parse_and_eval("isl_printer_to_str(isl_"
                                             + str(self.type)
					     + "_get_ctx(" + value + "))")
                printer = gdb.parse_and_eval("isl_printer_print_"
                                             + str(self.type) + "("
                                             + str(printer) + ", "
                                             + value + ")")
                string = gdb.parse_and_eval("(char*)isl_printer_get_str("
                                            + str(printer) + ")")
                gdb.parse_and_eval("isl_printer_free(" + str(printer) + ")")
                return string

	def display_hint (self):
                return 'string'

class IslIntPrinter:
	"""Print an isl_int """
	def __init__ (self, val):
                self.val = val

	def to_string (self):
                # Cast val to a void pointer to stop gdb using this pretty
                # printer for the pointer which would lead to an infinite loop.
                void_ptr = gdb.lookup_type('void').pointer()
                value = str(self.val.cast(void_ptr))

                context = gdb.parse_and_eval("isl_ctx_alloc()")
                printer = gdb.parse_and_eval("isl_printer_to_str("
                                             + str(context) + ")")
                printer = gdb.parse_and_eval("isl_printer_print_isl_int("
                                             + str(printer) + ", "
                                             + value + ")")
                string = gdb.parse_and_eval("(char*)isl_printer_get_str("
                                            + str(printer) + ")")
                gdb.parse_and_eval("isl_printer_free(" + str(printer) + ")")
                gdb.parse_and_eval("isl_ctx_free(" + str(context) + ")")
                return string

	def display_hint (self):
                return 'string'

class IslPrintCommand (gdb.Command):
        """Print an isl value."""
        def __init__ (self):
                super (IslPrintCommand, self).__init__ ("islprint",
                                                        gdb.COMMAND_OBSCURE)
        def invoke (self, arg, from_tty):
                arg = gdb.parse_and_eval(arg);
                printer = str_lookup_function(arg)

                if printer == None:
                        print "No isl printer for this type"
                        return

                print printer.to_string()

IslPrintCommand()

def str_lookup_function (val):
	if val.type.code != gdb.TYPE_CODE_PTR:
		if str(val.type) == "isl_int":
			return IslIntPrinter(val)
                else:
                        return None

	lookup_tag = val.type.target()
	regex = re.compile ("^isl_(.*)$")

	if lookup_tag == None:
		return None

	m = regex.match (str(lookup_tag))

	if m:
                # Those types of printers defined in isl.
                if m.group(1) in ["basic_set", "set", "union_set", "basic_map",
                                  "map", "union_map", "qpolynomial",
                                  "pw_qpolynomial", "pw_qpolynomial_fold",
                                  "union_pw_qpolynomial",
                                  "union_pw_qpolynomial_fold"]:
                        return IslObjectPrinter(val, m.group(1))
        return None

# Do not register the pretty printer.
# gdb.current_objfile().pretty_printers.append(str_lookup_function)



import json
from isl import *

class Scop:
  def __init__(self, filename):
    f = open(filename, 'r')
    self.json = json.load(f)
    return

  def __str__(self):
    return json.dumps(self.json, indent=2)

  def __repr__(self):
    return str(self)

  @property
  def statements(self):
    return self.json['statements']

class Transforms:
  """
  Create a map that interchanges two dimensions 'A' and 'B'

  numberDimensions: The overall number of dimensions
  dimensionA: The dimension of dimension 'A'
  dimensionB: The dimension of dimension 'B'

  getInterchange(2, 0, 1):
  {[d0, d1] -> [d1, d0]}
  """
  @staticmethod
  def getInterchange(numberDimensions, dimensionA, dimensionB):

    dims = ['d' + str(i) for i in range(numberDimensions)]
    dimString = ",".join(dims)

    changedDims = dims
    first = dims[dimensionA]
    second = dims[dimensionB]
    changedDims[dimensionA] = second
    changedDims[dimensionB] = first
    changedDimString = ",".join(changedDims)

    return Map("{[%s] -> [%s]}" % (dimString, changedDimString))

  """
  Create a map that strip mines one dimension

  numberDimensions: The overall number of dimensions
  stripMineDim: The dimension to strip mine
  factor: The strip mining factor

  getStripMine(2, 1, 64):
  {[d0, d1] -> [d0, o, d1] : o % 64 = 0 and o <= d1 <= d1 + 63}
  """
  @staticmethod
  def getStripMine(numberDimensions, stripMineDim, factor):

    dims = ['d' + str(i) for i in range(numberDimensions)]
    dimString = ",".join(dims)

    changedDims = dims
    smd = dims[stripMineDim]
    changedDims[stripMineDim] = "o,%s" % smd
    changedDimString = ",".join(changedDims)
    string = "{[%s] -> [%s]: o %% %i = 0 and o <= %s <= o + %i}" % \
          (dimString, changedDimString, factor, smd, factor - 1)
    return Map(string)



#!/usr/bin/python
import argparse, isl, os
import json

def printDomain(scop):

  domain = isl.USet('{}')

  for statement in scop['statements']:
    domain = domain.union(isl.USet(statement['domain']))

  print "D :=",
  print str(domain) + ";"

def printAccesses(scop):

  read = isl.UMap('{}')

  for statement in scop['statements']:
    for access in statement['accesses']:
      if access['kind'] == 'read':
        read = read.union(isl.UMap(access['relation']))

  print "R :=",
  print str(read) + ";"

  write = isl.UMap('{}')

  for statement in scop['statements']:
    for access in statement['accesses']:
      if access['kind'] == 'write':
        write = write.union(isl.UMap(access['relation']))

  print "W :=",
  print str(write) + ";"

def printSchedule(scop):

  schedule = isl.UMap('{}')

  for statement in scop['statements']:
    schedule = schedule.union(isl.UMap(statement['schedule']))

  print "S :=",
  print str(schedule) + ";"

def __main__():
  description = 'Translate JSCoP into iscc input'
  parser = argparse.ArgumentParser(description)
  parser.add_argument('inputFile', metavar='N', type=file,
                      help='The JSCoP file')

  args = parser.parse_args()
  inputFile = args.inputFile
  scop = json.load(inputFile)

  printDomain(scop)
  printAccesses(scop)
  printSchedule(scop)

  print 'R := R * D;'
  print 'W := W * D;'
  print 'Dep := (last W before R under S)[0];'
  print 'schedule D respecting Dep minimizing Dep;'


__main__()




#! /usr/bin/env python3
# -*- coding: UTF-8 -*-

# Polly/LLVM update_check.py
# Update lit FileCheck files by replacing the 'CHECK:' lines by the actual output of the 'RUN:' command.

import argparse
import os
import subprocess
import shlex
import re


polly_src_dir = '''@POLLY_SOURCE_DIR@'''
polly_lib_dir = '''@POLLY_LIB_DIR@'''
shlibext = '''@LLVM_SHLIBEXT@'''
llvm_tools_dir = '''@LLVM_TOOLS_DIR@'''
link_polly_into_tools = not '''@LINK_POLLY_INTO_TOOLS@'''.lower() in {'','0','n','no','off','false','notfound','link_polly_into_tools-notfound'}

runre = re.compile(r'\s*\;\s*RUN\s*\:(?P<tool>.*)')
filecheckre = re.compile(r'\s*(?P<tool>.*)\|\s*(?P<filecheck>FileCheck\s[^|]*)')
emptyline = re.compile(r'\s*(\;\s*)?')
commentline = re.compile(r'\s*(\;.*)?')


def ltrim_emptylines(lines,meta=None):
    while len(lines) and emptyline.fullmatch(lines[0]):
        del lines[0]
        if meta is not None:
            del meta[0]


def rtrim_emptylines(lines):
    while len(lines) and emptyline.fullmatch(lines[-1]):
        del lines[-1]


def trim_emptylines(lines):
    ltrim_emptylines(lines)
    rtrim_emptylines(lines)


def complete_exename(path, filename):
    complpath = os.path.join(path, filename)
    if os.path.isfile(complpath):
        return complpath
    elif os.path.isfile(complpath + '.exe'):
        return complpath + '.exe'
    return filename


def indention(line):
    for i,c in enumerate(line):
        if c != ' ' and c != '\t':
            return i
    return None


def common_indent(lines):
    indentions = (indention(line) for line in lines)
    indentions = (indent for indent in indentions if indent is not None)
    return min(indentions,default=0)


funcre = re.compile(r'^    Function: \S*$')
regionre = re.compile(r'^    Region: \S*$')
depthre = re.compile(r'^    Max Loop Depth: .*')
paramre = re.compile(r'    [0-9a-z-A-Z_]+\: .*')

def classyfier1(lines):
    i = iter(lines)
    line = i.__next__()
    while True:
        if line.startswith("Printing analysis 'Polly - Calculate dependences' for region: "):
            yield {'PrintingDependenceInfo'}
        elif line.startswith("remark: "):
            yield {'Remark'}
        elif funcre.fullmatch(line):
            yield {'Function'}
        elif regionre.fullmatch(line):
            yield  { 'Region'}
        elif depthre.fullmatch(line):
            yield  {'MaxLoopDepth'}
        elif line == '    Invariant Accesses: {':
            while True:
                yield { 'InvariantAccesses'}
                if line == '    }':
                    break
                line = i.__next__()
        elif line == '    Context:':
            yield  {'Context'}
            line = i.__next__()
            yield  {'Context'}
        elif line == '    Assumed Context:':
            yield  {'AssumedContext'}
            line = i.__next__()
            yield  {'AssumedContext'}
        elif line == '    Invalid Context:':
            yield  {'InvalidContext'}
            line = i.__next__()
            yield  {'InvalidContext'}
        elif line == '    Boundary Context:':
            yield  {'BoundaryContext'}
            line = i.__next__()
            yield  {'BoundaryContext'}
            line = i.__next__()
            while paramre.fullmatch(line):
                yield  {'Param'}
                line = i.__next__()
            continue
        elif line == '    Arrays {':
            while True:
                yield  {'Arrays'}
                if line == '    }':
                    break
                line = i.__next__()
        elif line == '    Arrays (Bounds as pw_affs) {':
            while True:
                yield  {'PwAffArrays'}
                if line == '    }':
                    break
                line = i.__next__()
        elif line.startswith('    Alias Groups ('):
            while True:
                yield  {'AliasGroups'}
                line = i.__next__()
                if not line.startswith('        '):
                    break
            continue
        elif line == '    Statements {':
            while True:
                yield  {'Statements'}
                if line == '    }':
                    break
                line = i.__next__()
        elif line == '    RAW dependences:':
            yield {'RAWDep','BasicDep','Dep','DepInfo'}
            line = i.__next__()
            while line.startswith('        '):
                yield  {'RAWDep','BasicDep','Dep','DepInfo'}
                line = i.__next__()
            continue
        elif line == '    WAR dependences:':
            yield {'WARDep','BasicDep','Dep','DepInfo'}
            line = i.__next__()
            while line.startswith('        '):
                yield  {'WARDep','BasicDep','Dep','DepInfo'}
                line = i.__next__()
            continue
        elif line == '    WAW dependences:':
            yield {'WAWDep','BasicDep','Dep','DepInfo'}
            line = i.__next__()
            while line.startswith('        '):
                yield  {'WAWDep','BasicDep','Dep','DepInfo'}
                line = i.__next__()
            continue
        elif line == '    Reduction dependences:':
            yield {'RedDep','Dep','DepInfo'}
            line = i.__next__()
            while line.startswith('        '):
                yield  {'RedDep','Dep','DepInfo'}
                line = i.__next__()
            continue
        elif line == '    Transitive closure of reduction dependences:':
            yield {'TransitiveClosureDep','DepInfo'}
            line = i.__next__()
            while line.startswith('        '):
                yield  {'TransitiveClosureDep','DepInfo'}
                line = i.__next__()
            continue
        elif line.startswith("New access function '"):
            yield {'NewAccessFunction'}
        elif line == 'Schedule before flattening {':
            while True:
                yield  {'ScheduleBeforeFlattening'}
                if line == '}':
                    break
                line = i.__next__()
        elif line == 'Schedule after flattening {':
            while True:
                yield  {'ScheduleAfterFlattening'}
                if line == '}':
                    break
                line = i.__next__()
        else:
            yield set()
        line = i.__next__()


def classyfier2(lines):
    i = iter(lines)
    line = i.__next__()
    while True:
        if funcre.fullmatch(line):
            while line.startswith('    '):
                yield  {'FunctionDetail'}
                line = i.__next__()
            continue
        elif line.startswith("Printing analysis 'Polly - Generate an AST from the SCoP (isl)' for region: "):
            yield {'PrintingIslAst'}
            line = i.__next__()
            while not line.startswith('Printing analysis'):
                yield  {'AstDetail'}
                line = i.__next__()
            continue
        else:
            yield set()
        line = i.__next__()


replrepl = {'{{':'{{[{][{]}}','}}': '{{[}][}]}}', '[[':'{{\[\[}}',']]': '{{\]\]}}'}
replre = re.compile('|'.join(re.escape(k) for k in replrepl.keys()))

def main():
    parser = argparse.ArgumentParser(description="Update CHECK lines")
    parser.add_argument('testfile',help="File to update (absolute or relative to --testdir)")
    parser.add_argument('--check-style',choices=['CHECK','CHECK-NEXT'],default='CHECK-NEXT',help="What kind of checks lines to generate")
    parser.add_argument('--check-position',choices=['end','before-content','autodetect'],default='autodetect',help="Where to add the CHECK lines into the file; 'autodetect' searches for the first 'CHECK' line ind inserts it there")
    parser.add_argument('--check-include',action='append',default=[], help="What parts of the output lines to check; use syntax 'CHECK=include' to apply to one CHECK-prefix only (by default, everything)")
    parser.add_argument('--check-label-include',action='append',default=[],help="Use CHECK-LABEL for these includes")
    parser.add_argument('--check-part-newline',action='store_true',help="Add empty line between different check parts")
    parser.add_argument('--prefix-only',action='append',default=None,help="Update only these prefixes (default: all)")
    parser.add_argument('--bindir',help="Location of the opt program")
    parser.add_argument('--testdir',help="Root dir for unit tests")
    parser.add_argument('--inplace','-i',action='store_true',help="Replace input file")
    parser.add_argument('--output','-o',help="Write changed input to this file")
    known = parser.parse_args()

    if not known.inplace and known.output is None:
        print("Must specify what to do with output (--output or --inplace)")
        exit(1)
    if known.inplace and known.output is not None:
        print("--inplace and --output are mutually exclusive")
        exit(1)

    outfile = known.output

    filecheckparser = argparse.ArgumentParser(add_help=False)
    filecheckparser.add_argument('-check-prefix','--check-prefix',default='CHECK')

    filename = known.testfile
    for dir in ['.', known.testdir, os.path.join(polly_src_dir,'test'), polly_src_dir]:
        if not dir:
            continue
        testfilename = os.path.join(dir,filename)
        if os.path.isfile(testfilename):
            filename = testfilename
            break

    if known.inplace:
        outfile = filename

    allchecklines = []
    checkprefixes = []

    with open(filename, 'r') as file:
        oldlines = [line.rstrip('\r\n') for line in file.readlines()]

    runlines = []
    for line in oldlines:
        m = runre.match(line)
        if m:
            runlines.append(m.group('tool'))

    continuation = ''
    newrunlines = []
    for line in runlines:
        if line.endswith('\\'):
            continuation += line[:-2] + ' '
        else:
            newrunlines.append(continuation + line)
            continuation = ''
    if continuation:
        newrunlines.append(continuation)

    for line in newrunlines:
        m = filecheckre.match(line)
        if not m:
            continue

        tool, filecheck = m.group('tool', 'filecheck')
        filecheck = shlex.split(filecheck)
        tool = shlex.split(tool)
        if known.bindir is not None:
            tool[0] = complete_exename(known.bindir, tool[0])
        if os.path.isdir(llvm_tools_dir):
            tool[0] = complete_exename(llvm_tools_dir, tool[0])
        check_prefix = filecheckparser.parse_known_args(filecheck)[0].check_prefix
        if known.prefix_only is not None and not check_prefix in known.prefix_only:
            continue
        if check_prefix in checkprefixes:
            continue
        checkprefixes.append(check_prefix)

        newtool = []
        optstderr = None
        for toolarg in tool:
            toolarg = toolarg.replace('%s', filename)
            toolarg = toolarg.replace('%S', os.path.dirname(filename))
            if toolarg == '%loadPolly':
                if not link_polly_into_tools:
                    newtool += ['-load',os.path.join(polly_lib_dir,'LLVMPolly' + shlibext)]
                newtool.append('-polly-process-unprofitable')
                newtool.append('-polly-remarks-minimal')
            elif toolarg == '2>&1':
                optstderr = subprocess.STDOUT
            else:
                newtool.append(toolarg)
        tool = newtool

        inpfile = None
        i = 1
        while i <  len(tool):
            if tool[i] == '<':
                inpfile = tool[i + 1]
                del tool[i:i + 2]
                continue
            i += 1
        if inpfile:
            with open(inpfile) as inp:
                retlines = subprocess.check_output(tool,universal_newlines=True,stdin=inp,stderr=optstderr)
        else:
            retlines = subprocess.check_output(tool,universal_newlines=True,stderr=optstderr)
        retlines = [line.replace('\t', '    ') for line in retlines.splitlines()]
        check_include = []
        for checkme in known.check_include + known.check_label_include:
            parts = checkme.split('=')
            if len(parts) == 2:
                if parts[0] == check_prefix:
                    check_include.append(parts[1])
            else:
                check_include.append(checkme)

        if check_include:
            filtered_retlines = []
            classified_retlines = []
            lastmatch = None
            for line,kind in ((line,class1.union(class2)) for line,class1,class2 in zip(retlines,classyfier1(retlines), classyfier2(retlines))):
                match = kind.intersection(check_include)
                if match:
                    if lastmatch != match:
                        filtered_retlines.append('')
                        classified_retlines.append({'Separator'})
                    filtered_retlines.append(line)
                    classified_retlines.append(kind)
                lastmatch = match

            retlines = filtered_retlines
        else:
            classified_retlines = (set() for line in retlines)

        rtrim_emptylines(retlines)
        ltrim_emptylines(retlines,classified_retlines)
        retlines = [replre.sub(lambda m: replrepl[m.group(0)], line) for line in retlines]
        indent = common_indent(retlines)
        retlines = [line[indent:] for line in retlines]
        checklines = []
        previous_was_empty = True
        for line,kind in zip(retlines,classified_retlines):
            if line:
                if known.check_style == 'CHECK' and known.check_label_include:
                    if not kind.isdisjoint(known.check_label_include):
                        checklines.append('; ' + check_prefix + '-LABEL: ' + line)
                    else:
                        checklines.append('; ' + check_prefix + ':       ' + line)
                elif known.check_style == 'CHECK':
                    checklines.append('; ' + check_prefix + ': ' + line)
                elif known.check_label_include and known.check_label_include:
                    if not kind.isdisjoint(known.check_label_include):
                        checklines.append('; ' + check_prefix + '-LABEL: ' + line)
                    elif previous_was_empty:
                        checklines.append('; ' + check_prefix + ':       ' + line)
                    else:
                        checklines.append('; ' + check_prefix + '-NEXT:  ' + line)
                else:
                    if previous_was_empty:
                        checklines.append('; ' + check_prefix + ':      ' + line)
                    else:
                        checklines.append('; ' + check_prefix + '-NEXT: ' + line)
                previous_was_empty = False
            else:
                if not 'Separator' in kind or known.check_part_newline:
                    checklines.append(';')
                previous_was_empty = True
        allchecklines.append(checklines)

    if not checkprefixes:
        return

    checkre = re.compile(r'^\s*\;\s*(' + '|'.join([re.escape(s) for s in checkprefixes]) + ')(\-NEXT|\-DAG|\-NOT|\-LABEL|\-SAME)?\s*\:')
    firstcheckline = None
    firstnoncommentline = None
    headerlines = []
    newlines = []
    uptonowlines = []
    emptylines = []
    lastwascheck = False
    for line in oldlines:
        if checkre.match(line):
            if firstcheckline is None:
                firstcheckline = len(newlines) + len(emptylines)
            if not lastwascheck:
                uptonowlines += emptylines
            emptylines = []
            lastwascheck = True
        elif emptyline.fullmatch(line):
            emptylines.append(line)
        else:
            newlines += uptonowlines
            newlines += emptylines
            newlines.append(line)
            emptylines = []
            uptonowlines = []
            lastwascheck = False

    for i,line in enumerate(newlines):
        if not commentline.fullmatch(line):
            firstnoncommentline = i
            break

    with open(outfile,'w',newline='') as file:
        def writelines(lines):
            for line in lines:
                file.write(line)
                file.write('\n')

        if firstcheckline is not None and known.check_position == 'autodetect':
            writelines(newlines[:firstcheckline])
            writelines(uptonowlines)
            for i,checklines in enumerate(allchecklines):
                if i != 0:
                    file.write('\n')
                writelines(checklines)
            writelines(newlines[firstcheckline:])
            writelines(emptylines)
        elif firstnoncommentline is not None and known.check_position == 'before-content':
            headerlines = newlines[:firstnoncommentline]
            rtrim_emptylines(headerlines)
            contentlines = newlines[firstnoncommentline:]
            ltrim_emptylines(contentlines)

            writelines(headerlines)
            for checklines in allchecklines:
                file.write('\n')
                writelines(checklines)
            file.write('\n')
            writelines(contentlines)
            writelines(uptonowlines)
            writelines(emptylines)
        else:
            writelines(newlines)
            rtrim_emptylines(newlines)
            for checklines in allchecklines:
                file.write('\n\n')
                writelines(checklines)


if __name__ == '__main__':
    main()
